      subroutine SLhpf_cgels_v(A, B, trans0, info0)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     Written by C. A. Papadopoulos, University of Tennessee, Knoxville
!     Adapted from SLhpf routines written by L. Susan Blackford,
!     and R. Clint Whaley
!     University of Tennessee, Knoxville
!     September 1, 1998
!
      implicit none
      include 'pblas.h'
      include 'misc.h'
      complex(SP), intent(inout), dimension(:,:) :: A
      complex(SP), intent(inout), dimension(:) :: B
      character(LEN=1), optional, intent(in) :: trans0
      integer, optional, intent(out) :: info0
!HPF$ INHERIT A, B
      character(len=1) :: trans
      logical :: notrans
      integer :: icontxt, nprow, npcol, myrow, mycol
      integer :: i, j, P, Q, maxQ, info, mA, nA, mbA, nbA, rsrcA, csrcA
      integer :: mB, nB, mbB, nbB, rsrcB, csrcB

      integer, dimension(DLEN_,NUMBER_OF_PROCESSORS()) :: descA, descB
!HPF$ DISTRIBUTE (*,block) :: descA, descB

      INTERFACE

      subroutine SLhpf_cdescset3_v(A, descA, B, descB, C, descC, P0, Q0,  &
     &                           info0)
         include 'pblas.h'
         include 'misc.h'
         complex(SP), dimension(:,:), intent(in) :: A
         integer, dimension(DLEN_,NUMBER_OF_PROCESSORS()),              &
     &            intent(out) :: descA
         complex(SP), dimension(:), optional, intent(in) :: B
         complex(SP), dimension(:,:), optional, intent(in) :: C
         integer, optional, dimension(DLEN_,NUMBER_OF_PROCESSORS()),    &
     &            intent(out) :: descB, descC
         integer, optional, intent(out) :: P0, Q0, info0
!HPF$    INHERIT A, B, C
!HPF$    DISTRIBUTE(*,block) :: descA, descB, descC
      end subroutine SLhpf_cdescset3_v

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_DescQuery(desc, M, N, MB, NB, RSRC, CSRC)
         include 'pblas.h'
         integer, intent(in) :: desc(:,:)
!HPF$    DISTRIBUTE(*,block) :: desc
         integer, optional, intent(out) :: M, N, MB, NB, RSRC, CSRC
      end subroutine SLhpf_DescQuery

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_cgels1_v(trans, m, n, nrhs, A, ia, ja, descA, B,   &
     &                        ib, jb, descB, info)
         include 'pblas.h'
         include 'misc.h'
         character(LEN=1), intent(in) :: trans
         integer, intent(in) :: m, n, nrhs, ia, ja, ib, jb
         integer, dimension(:,:), intent(inout) :: descA, descB
         integer, intent(out) :: info
         complex(SP), intent(inout), dimension(:,:) :: A
         complex(SP), intent(inout), dimension(:) :: B
!HPF$    INHERIT A, B
!HPF$    DISTRIBUTE(*,block) :: descA, descB
      end subroutine SLhpf_cgels1_v

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_cgels2_v(P, Q, mbA, nbA, mbB, nbB, trans, m, n,    &
     &                          nrhs, A, ia, ja, descA, B, ib, jb,      &
     &                          descB, info)
         include 'pblas.h'
         include 'misc.h'
         character(LEN=1), intent(in) :: trans
         integer, intent(in) :: P, Q, mbA, nbA, mbB, nbB, m, n, nrhs,   &
     &                          ia, ja, ib, jb
         integer, dimension(:,:), intent(inout) :: descA, descB
         integer, intent(out) :: info
         complex(SP), intent(inout), dimension(:,:) :: A
         complex(SP), intent(inout), dimension(:) :: B
!HPF$    PROCESSORS PROC(P,Q)
!HPF$ DISTRIBUTE A(CYCLIC(mbA), CYCLIC(nbA)) ONTO PROC
!HPF$ ALIGN B(I) with A(I,1)
! due to a pghpf compiler bug cannot align with n which would
! optimize performance
! !HPF$ ALIGN B(I) with A(I,n)
!HPF$ DISTRIBUTE(*,block) :: descA, descB
      end subroutine SLhpf_cgels2_v

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_get_gridshape(P, Q, maxP, maxQ)
         integer, intent(out) :: P, Q
         integer, optional, intent(in) :: maxP, maxQ
      end subroutine SLhpf_get_gridshape

      extrinsic(hpf_local) subroutine SLhpf_abort(errno)
         integer, intent(in) :: errno
      end subroutine SLhpf_abort

      END INTERFACE

      info = 0
      if ( present(trans0) ) then
         trans = trans0
         notrans = ( trans.eq.'n' .or. trans.eq.'N' )
      else
         notrans = .true.
         trans = 'N'
      end if

      call SLhpf_cdescset3_v(A, descA, B, descB, P0=P, Q0=Q, info0=info)
      call SLhpf_DescQuery(descA, mA, nA, mbA, nbA, rsrcA, csrcA)
      call SLhpf_DescQuery(descB, mB, nB, mbB, nbB, rsrcB, csrcB)
!
!     If info shows error in descriptor, shift it to correct number
!
      if (info .lt. 0) then
         if (info .lt. -200) then
            info = -1200 - MOD(-info, 100)
         else
            info = -800 - MOD(-info, 100)
         end if
      end if
!
!     Check for illegal size subsection.  If we have one, it is a fatal error.
!
      if ( mB .ne. max(mA, nA) ) then
         if ( present(info0) ) then
            info0 = -9
         else
            print*,'SIZE(B,1) SHOULD be MAX( SIZE(A,1), SIZE(A,2) )'
            print*,'PCGELS_V exited with error code ',info
            call SLhpf_abort(info)
            stop
         end if
         return
      end if
!
!     These represent alignment restrictions in present version pcgels
!
      if (mA .ge. nA) then
         if (rsrcA .ne. rsrcB) then
            descB(RSRC_,:) = descA(RSRC_,:)
            if (info .eq. 0) info = -(1200 + RSRC_)
         end if
         if (mbA .ne. mbB) then
            descB(MB_,:) = descA(MB_,:)
            if (info .eq. 0) info = -(1200 + MB_)
         end if
      else
         if (nbA .ne. mbB) then
            descB(MB_,:) = descA(NB_,:)
            if (info .eq. 0) info = -(1200 + MB_)
         end if
      end if
!
      if (info .eq. 0) then
!
!     No redistribution needed !!
!
         call SLhpf_cgels1_v(trans, mA, nA, nB, A, 1, 1, descA, B, 1, 1,   &
     &                     descB, info)
!
      else
!
!     Redistribution needed !!
!
         if (REDIST) then
            write(*,*) "WARNING:  redistribution is necessary"
         end if
!
         call SLhpf_DescQuery(descA, mA, nA, mbA, nbA)
         call SLhpf_DescQuery(descB, mB, nB, mbB, nbB)
         info = 0
         call SLhpf_get_gridshape(P, Q)
         call SLhpf_cgels2_v(P, Q, mbA, nbA, mbB, nbB, trans, mA, nA,     &
     &                     nB, A, 1, 1, descA, B, 1, 1, descB, info)
      end if

      if ( present(info0) ) then
         info0 = info
      else if (info .ne. 0) then
         print*,'PCGELS exited with error code ',info
         call SLhpf_abort(info)
         stop
      end if

      return
      end

