      subroutine SLhpf_cgemm(A, B, C, transA0, transB0, alpha0, beta0)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     September 1, 1998
!     University of Tennessee, Knoxville
!     Code contributors: R. Clint Whaley
!     Last modified by: C. A. Papadopoulos
!
      implicit none
      include 'pblas.h'
      include 'misc.h'
      complex(SP), intent(in), dimension(:,:) :: A, B
      complex(SP), intent(inout), dimension(:,:) :: C
      character(LEN=1), optional, intent(in) :: transA0, transB0
      complex(SP), optional, intent(in) :: alpha0, beta0
!HPF$ INHERIT A, B, C

      character(LEN=1) :: transA, transB
      LOGICAL :: NOTRANSA, NOTRANSB
      integer :: P, Q, csrcA, csrcB, csrcC, info, mA, mAop, mB, mBop,   &
     &           mC, mbA, mbAop, mbB, mbBop, mbC, nA, nAop, nB, nBop,   &
     &           nC, nbA, nbAop, nbB, nbBop, nbC, rsrcA, rsrcB, rsrcC
      complex(SP) :: alpha, beta
      integer, dimension(DLEN_,NUMBER_OF_PROCESSORS()) :: descA, descB, &
     &                                                    descC
!HPF$ DISTRIBUTE (*,block) :: descA, descB
      INTERFACE

      subroutine SLhpf_cdescset3(A, descA, B, descB, C, descC, P0, Q0,  &
     &                           info0)
         include 'pblas.h'
         include 'misc.h'
         complex(SP), dimension(:,:), intent(in) :: A
         integer, dimension(DLEN_,NUMBER_OF_PROCESSORS()),              &
     &            intent(out) :: descA
         complex(SP), dimension(:,:), optional, intent(in) :: B, C
         integer, optional, dimension(DLEN_,NUMBER_OF_PROCESSORS()),    &
     &            intent(out) :: descB, descC
         integer, optional, intent(out) :: P0, Q0, info0
!HPF$    INHERIT A, B, C
!HPF$    DISTRIBUTE(*,block) :: descA, descB, descC
      end subroutine SLhpf_cdescset3

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_DescQuery(desc, M, N, MB, NB, RSRC, CSRC)
         include 'pblas.h'
         integer, intent(in) :: desc(:,:)
!HPF$    DISTRIBUTE(*,block) :: desc
         integer, optional, intent(out) :: M, N, MB, NB, RSRC, CSRC
      end subroutine SLhpf_DescQuery

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_cgemm1(transA, transB, m, n, k, alpha, A,        &
     &                        ia, ja, descA, B, ib, jb, descB, beta,    &
     &                        C, ic, jc, descC)
         include 'pblas.h'
         include 'misc.h'
         character(LEN=1), intent(in) :: transA, transB
         integer, intent(in) :: m, n, k, ia, ja, ib, jb, ic, jc
         integer, dimension(:,:), intent(inout) :: descA, descB, descC
         complex(SP), intent(in) :: alpha, beta
         complex(SP), intent(in), dimension(:,:) :: A, B
         complex(SP), intent(inout), dimension(:,:) :: C
!HPF$    INHERIT A, B, C
!HPF$    DISTRIBUTE(*,block) :: descA, descB, descC
      end subroutine SLhpf_cgemm1

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_cgemm2(P, Q, mbA, nbA, mbB, nbB, mbC, nbC,       &
     &                        transA, transB, m, n, k, alpha, A,        &
     &                        ia, ja, descA, B, ib, jb, descB, beta,    &
     &                        C, ic, jc, descC)
         include 'pblas.h'
         include 'misc.h'
         character(LEN=1), intent(in) :: transA, transB
         integer, intent(in) :: P, Q, mbA, nbA, mbB, nbB, mbC, nbC,     &
     &                          m, n, k, ia, ja, ib, jb, ic, jc
         integer, dimension(:,:), intent(inout) :: descA, descB, descC
         complex(SP), intent(in) :: alpha, beta
         complex(SP), intent(in), dimension(:,:) :: A, B
         complex(SP), intent(inout), dimension(:,:) :: C
!HPF$    PROCESSORS PROC(P,Q)
!HPF$ DISTRIBUTE A(CYCLIC(mbA), CYCLIC(nbA)) ONTO PROC
!HPF$ DISTRIBUTE B(CYCLIC(mbB), CYCLIC(nbB)) ONTO PROC
!HPF$ DISTRIBUTE C(CYCLIC(mbC), CYCLIC(nbC)) ONTO PROC
!HPF$    DISTRIBUTE(*,block) :: descA, descB, descC
      end subroutine SLhpf_cgemm2

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_get_gridshape(P, Q, maxP, maxQ)
         integer, intent(out) :: P, Q
         integer, optional, intent(in) :: maxP, maxQ
      end subroutine SLhpf_get_gridshape

      extrinsic(hpf_local) subroutine SLhpf_abort(errno)
         integer, intent(in) :: errno
      end subroutine SLhpf_abort

      END INTERFACE
!
!     Set implicit parameters
!
      NOTRANSA = .TRUE.
      NOTRANSB = .TRUE.
      if ( present(alpha0) ) then
         alpha = alpha0
      else
         alpha = (1.0E0, 0.0E0)
      end if
      if ( present(beta0) ) then
         beta = beta0
      else
         beta = (0.0E0, 0.0E0)
      end if
      if ( present(transA0) ) then
         transA = transA0
         NOTRANSA = ( transA.eq.'n' .or. transA.eq.'N' )
      else
         transA = 'N'
      end if
      if ( present(transB0) ) then
         transB = transB0
         NOTRANSB = ( transB.eq.'n' .or. transB.eq.'N' )
      else
         transB = 'N'
      end if
!
!     Fetch matrix information
!
      call SLhpf_cdescset3(A, descA, B, descB, C, descC, P, Q, info)
      call SLhpf_DescQuery(descA, mA, nA, mbA, nbA, rsrcA, csrcA)
      call SLhpf_DescQuery(descB, mB, nB, mbB, nbB, rsrcB, csrcB)
      call SLhpf_DescQuery(descC, mC, nC, mbC, nbC, rsrcC, csrcC)
      if (NOTRANSA) then
         mAop = mA
         nAop = nA
         mbAop = mbA
         nbAop = nbA
      else
         mAop = nA
         nAop = mA
         mbAop = nbA
         nbAop = mbA
      end if
      if (NOTRANSB) then
         mBop = mB
         nBop = nB
         mbBop = mbB
         nbBop = nbB
      else
         mBop = nB
         nBop = mB
         mbBop = nbB
         nbBop = mbB
      end if
!
!     If info shows error in descriptor, shift it to correct number
!
      if (info .lt. 0) then
         if (info .lt. -300) then
            info = -1900 - MOD(-info, 100)
         else if (info .lt. -200) then
            info = -1400 - MOD(-info, 100)
         else
            info = -1000 - MOD(-info, 100)
         end if
      end if
!
!     Check for illegal size subsection (terminal error)
!
      if ( (mC.ne.mAop) .or. (nC.ne.nBop) .or. (nAop.ne.mBop) ) then
         if (mC .ne. mAop) then
            info = -1
         else if (nC .ne. nBop) then
            info = -2
         else if (nAop .ne. mBop) then
            info = -3
         end if
         print*,'PCGEMM exited with error code ',info
         call SLhpf_abort(info)
         stop
      end if
!
!     These checks represent alignment restrictions in the present p_gemm
!
      if (mbC .ne. mbAop) then
         if (NOTRANSA) then
            info = -1000 - MB_
            descA(MB_,:) = descC(MB_,:)
         else
            info = -1000 - NB_
            descA(NB_,:) = descC(MB_,:)
         end if
      end if
      if (nbC .ne. nbBop) then
         if (NOTRANSB) then
            info = -1400 - NB_
            descA(NB_,:) = descC(NB_,:)
         else
            info = -1400 - MB_
            descA(MB_,:) = descC(NB_,:)
         end if
      end if
      if (mbBop .ne. nbAop) then
         if (NOTRANSB) then
            info = -1400 - MB_
            descB(MB_,:) = nbAop
         else
            info = -1400 - NB_
            descB(NB_,:) = nbAop
         end if
      end if
      if (NOTRANSA) then
         if (NOTRANSB) then
            if (rsrcC.ne.rsrcA) then
               info = -1000 - RSRC_
               descA(RSRC_,:) = descC(RSRC_,:)
            end if
            if (csrcC.ne.csrcB) then
               info = -1400 - CSRC_
               descB(CSRC_,:) = descC(CSRC_,:)
            end if
         else
            if (rsrcC.ne.rsrcA) then
               info = -1000 - RSRC_
               descA(RSRC_,:) = descC(RSRC_,:)
            end if
            if (csrcA.ne.csrcB) then
               info = -1400 - CSRC_
               descB(CSRC_,:) = descA(CSRC_,:)
            end if
         end if
      else
         if (NOTRANSB) then
            if (rsrcA.ne.rsrcB) then
               info = -1400 - RSRC_
               descB(RSRC_,:) = descA(RSRC_,:)
            end if
            if (csrcC.ne.csrcB) then
               info = -1400 - CSRC_
               descB(CSRC_,:) = descA(CSRC_,:)
            end if
         end if
      end if
!
      if (info .eq. 0) then
!
!     No redistribution needed !!
!
         call SLhpf_cgemm1(transA, transB, mC, nC, nAop, alpha, A,      &
     &                     1, 1, descA, B, 1, 1, descB, beta, C, 1, 1,  &
     &                     descC)
!
      else
!
!     Redistribution needed !!
!
         if (REDIST) then
            write(*,*) "WARNING:  redistribution is necessary"
         end if
         info = 0
!
!        Since we are redistributing, use optimal parameters all around
!
         call SLhpf_get_gridshape(P, Q)
         descA(MB_,:) = opt_NB
         descA(NB_,:) = opt_NB
         descA(RSRC_,:) = 0
         descA(CSRC_,:) = 0
         descB(MB_,:) = opt_NB
         descB(NB_,:) = opt_NB
         descB(RSRC_,:) = 0
         descB(CSRC_,:) = 0
         descC(MB_,:) = opt_NB
         descC(NB_,:) = opt_NB
         descC(RSRC_,:) = 0
         descC(CSRC_,:) = 0
         call SLhpf_cgemm2(P, Q, opt_NB, opt_NB, opt_NB, opt_NB,        &
     &                        opt_NB, opt_NB, transA, transB, mC, nC,   &
     &                        nAop, alpha, A, 1, 1, descA, B, 1, 1,     &
     &                        descB, beta, C, 1, 1, descC)
      endif

      return
      end
