      subroutine SLhpf_cgesv(A, B, ipiv, info0)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     September 1, 1998
!     University of Tennessee, Knoxville
!     Code contributors: R. Clint Whaley
!     Last modified by: C. A. Papadopoulos
!
      implicit none
      include 'pblas.h'
      include 'misc.h'
      complex(SP), intent(inout), dimension(:,:) :: A, B
      integer, optional, intent(out) :: ipiv(:), info0
!HPF$ INHERIT A, B
!HPF$ ALIGN IPIV(:) WITH A(:,*)

      integer :: P, Q, csrcA, csrcB, i, info, j, mA, mB, maxQ, mbA, mbB,&
     &           nA, nB, nbA, nbB, rsrcA, rsrcB
      integer, dimension(DLEN_,NUMBER_OF_PROCESSORS()) :: descA, descB
!HPF$ DISTRIBUTE (*,block) :: descA, descB

      INTERFACE

      subroutine SLhpf_cdescset3(A, descA, B, descB, C, descC, P0, Q0,  &
     &                           info0)
         include 'pblas.h'
         include 'misc.h'
         complex(SP), dimension(:,:), intent(in) :: A
         integer, dimension(DLEN_,NUMBER_OF_PROCESSORS()),              &
     &            intent(out) :: descA
         complex(SP), dimension(:,:), optional, intent(in) :: B, C
         integer, optional, dimension(DLEN_,NUMBER_OF_PROCESSORS()),    &
     &            intent(out) :: descB, descC
         integer, optional, intent(out) :: P0, Q0, info0
!HPF$    INHERIT A, B, C
!HPF$    DISTRIBUTE(*,block) :: descA, descB, descC
      end subroutine SLhpf_cdescset3

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_DescQuery(desc, M, N, MB, NB, RSRC, CSRC)
         include 'pblas.h'
         integer, intent(in) :: desc(:,:)
!HPF$    DISTRIBUTE(*,block) :: desc
         integer, optional, intent(out) :: M, N, MB, NB, RSRC, CSRC
      end subroutine SLhpf_DescQuery

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_cgesv1(n, nrhs, A, ia, ja, descA, B, ib, jb,     &
     &                        descB, info, ipiv)
         include 'pblas.h'
         include 'misc.h'
         integer, intent(in) :: n, nrhs, ia, ja, ib, jb
         integer, dimension(:,:), intent(in) :: descA, descB
         integer, optional, intent(out) :: IPIV(:)
         integer, intent(out) :: info
         complex(SP), intent(inout), dimension(:,:) :: A, B
!HPF$    INHERIT A, B, IPIV
!HPF$    DISTRIBUTE(*,block) :: descA, descB
      end subroutine SLhpf_cgesv1

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_cgesv2(P, Q, mbA, nbA, mbB, nbB, n, nrhs, A,     &
     &                  ia, ja, descA, B, ib, jb, descB, info, ipiv)
         include 'pblas.h'
         include 'misc.h'
         integer, intent(in) :: P, Q, mbA, nbA, mbB, nbB, n, nrhs,      &
     &                          ia, ja, ib, jb
         integer, dimension(:,:), intent(inout) :: descA, descB
         integer, optional, intent(out) :: IPIV(:)
         integer, intent(out) :: info
         complex(SP), intent(inout), dimension(:,:) :: A, B
!HPF$    PROCESSORS PROC(P,Q)
!HPF$ DISTRIBUTE A(CYCLIC(mbA), CYCLIC(nbA)) ONTO PROC
!HPF$ DISTRIBUTE B(CYCLIC(mbB), CYCLIC(nbB)) ONTO PROC
!HPF$    ALIGN IPIV(:)  WITH A(:,*)
!HPF$    DISTRIBUTE(*,block) :: descA, descB
      end subroutine SLhpf_cgesv2

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_get_gridshape(P, Q, maxP, maxQ)
         integer, intent(out) :: P, Q
         integer, optional, intent(in) :: maxP, maxQ
      end subroutine SLhpf_get_gridshape

      extrinsic(hpf_local) subroutine SLhpf_abort(errno)
         integer, intent(in) :: errno
      end subroutine SLhpf_abort

      END INTERFACE

      info = 0

      call SLhpf_cdescset3(A, descA, B, descB, P0=P, Q0=Q, info0=info)
      call SLhpf_DescQuery(descA, mA, nA, mbA, nbA, rsrcA, csrcA)
      call SLhpf_DescQuery(descB, mB, nB, mbB, nbB, rsrcB, csrcB)
!
!     If info shows error in descriptor, shift it to correct number
!
      if (info .lt. 0) then
         if (info .lt. -200) then
            info = -1100 - MOD(-info, 100)
         else
            info = -600 - MOD(-info, 100)
         end if
      end if
!
!     Check for illegal size subsection
!
      if ( (mA.ne.nA) .or. (mA.ne.mB) .or. (size(ipiv).ne.mA) ) then
         if ( present(info0) ) then
            info0 = -1
         else
            print*,'PCGESV exited with error code ',info
            call SLhpf_abort(info)
            stop
         end if
         return
      end if
      if (mA .eq. 0) return
!
!     These represent alignment restrictions in present version pcgesv
!
      if (mbA .ne. nbA) then          ! A must have square blocks
         descA(MB_,:) = opt_NB
         descA(NB_,:) = opt_NB
         if (info .eq. 0) info = -600 - NB_
      end if
      if (mbA .ne. mbB) then          ! force same row blocking
         descB(MB_,:) = descA(MB_,:)
         if (info .eq. 0) info = -1100 - MB_
      end if
      if (rsrcA .ne. rsrcB) then      ! A and B must be row-aligned
         descB(RSRC_,:) = descA(RSRC_,:)
         if (info .eq. 0) info = -1100 - RSRC_
      end if
!
!     The following restriction gets around not knowing how to replicate
!     IPIV if A does not span the columns of the process grid
!
      maxQ = (nA + nbA-1) / nbA
      if ( maxQ .lt. Q ) then
         info = -600 - CTXT_
      end if
!
      if (info .eq. 0) then
!
!     No redistribution needed !!
!
         call SLhpf_cgesv1(mA, nB, A, 1, 1, descA, B, 1, 1, descB,      &
     &                     info, ipiv)
      else
!
!     Redistribution needed !!
!
         if (REDIST) then
            write(*,*) "WARNING:  redistribution is necessary"
         end if
!
         call SLhpf_DescQuery(descA, mA, nA, mbA, nbA)
         call SLhpf_DescQuery(descB, mB, nB, mbB, nbB)
         info = 0
         call SLhpf_get_gridshape(P, Q, maxQ=maxQ)
         call SLhpf_cgesv2(P, Q, mbA, nbA, mbB, nbB, mA, nB, A,         &
     &                        1, 1, descA, b, 1, 1, descB, info, ipiv)
      end if

      if ( present(info0) ) then
         info0 = info
      else if (info .ne. 0) then
         print*,'PCGESV exited with error code ',info
         call SLhpf_abort(info)
         stop
      end if

      return
      end


