      subroutine SLhpf_dtrsm(A, B, side0, uplo0, trans0, diag0, alpha0)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     September 1, 1998
!     University of Tennessee, Knoxville
!     Code contributors: R. Clint Whaley
!     Last modified by: C. A. Papadopoulos
!
      implicit none
      include 'pblas.h'
      include 'misc.h'
      real(DP), intent(in), dimension(:,:) :: A
      real(DP), intent(inout), dimension(:,:) :: B
      character(LEN=1), optional, intent(in) :: side0, uplo0, trans0,   &
     &                                          diag0
      real(DP), optional, intent(in) :: alpha0
!HPF$ INHERIT A, B

      character(LEN=1) :: side, uplo, transA, diag
      LOGICAL :: NOTRANSA, NOTRANSB
      integer :: K, M, N, P, Q, csrcA, csrcB, csrcC, info, mA, mB, mC,  &
     &           mbA, mbB, mbC, nA, nB, nC, nbA, nbB, nbC, rsrcA, rsrcB,&
     &           rsrcC
      real(DP) :: alpha, beta
      integer, dimension(DLEN_,NUMBER_OF_PROCESSORS()) :: descA, descB
!HPF$ DISTRIBUTE (*,block) :: descA, descB
      INTERFACE

      subroutine SLhpf_ddescset3(A, descA, B, descB, C, descC, P0, Q0,  &
     &                           info0)
         include 'pblas.h'
         include 'misc.h'
         real(DP), dimension(:,:), intent(in) :: A
         integer, dimension(DLEN_,NUMBER_OF_PROCESSORS()),              &
     &            intent(out) :: descA
         real(DP), dimension(:,:), optional, intent(in) :: B, C
         integer, optional, dimension(DLEN_,NUMBER_OF_PROCESSORS()),    &
     &            intent(out) :: descB, descC
         integer, optional, intent(out) :: P0, Q0, info0
!HPF$    INHERIT A, B, C
!HPF$    DISTRIBUTE(*,block) :: descA, descB, descC
      end subroutine SLhpf_ddescset3

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_DescQuery(desc, M, N, MB, NB, RSRC, CSRC)
         include 'pblas.h'
         integer, intent(in) :: desc(:,:)
!HPF$    DISTRIBUTE(*,block) :: desc
         integer, optional, intent(out) :: M, N, MB, NB, RSRC, CSRC
      end subroutine SLhpf_DescQuery

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_dtrsm1(side, uplo, transA, diag, m, n, alpha, A, &
     &                        ia, ja, descA, B, ib, jb, descB)
               include 'pblas.h'
         include 'misc.h'
         character(LEN=1), intent(in) :: side, uplo, transA, diag
         integer, intent(in) :: m, n, ia, ja, ib, jb
         integer, dimension(:,:), intent(inout) :: descA, descB
         real(DP), intent(in) :: alpha
         real(DP), intent(in), dimension(:,:) :: A
         real(DP), intent(inout), dimension(:,:) :: B
!HPF$ INHERIT A, B
!HPF$ DISTRIBUTE(*,block) :: descA, descB
      end subroutine SLhpf_dtrsm1

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_dtrsm2(P, Q, mbA, nbA, mbB, nbB,                 &
     &                        side, uplo, transA, diag, m, n, alpha,    &
     &                        A, ia, ja, descA, B, ib, jb, descB)
         include 'pblas.h'
         include 'misc.h'
         character(LEN=1), intent(in) :: side, uplo, transA, diag
         integer, intent(in) :: P, Q, mbA, nbA, mbB, nbB, m, n, ia, ja,    &
     &                       ib, jb
         integer, dimension(:,:), intent(inout) :: descA, descB
         real(DP), intent(in) :: alpha
         real(DP), intent(in), dimension(:,:) :: A
         real(DP), intent(inout), dimension(:,:) :: B
!HPF$ PROCESSORS PROC(P,Q)
!HPF$ DISTRIBUTE A(CYCLIC(mbA), CYCLIC(nbA)) ONTO PROC
!HPF$ DISTRIBUTE B(CYCLIC(mbB), CYCLIC(nbB)) ONTO PROC
!HPF$ DISTRIBUTE(*,block) :: descA, descB
      end subroutine SLhpf_dtrsm2

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_get_gridshape(P, Q, maxP, maxQ)
         integer, intent(out) :: P, Q
         integer, optional, intent(in) :: maxP, maxQ
      end subroutine SLhpf_get_gridshape

      extrinsic(hpf_local) subroutine SLhpf_abort(errno)
         integer, intent(in) :: errno
      end subroutine SLhpf_abort

      END INTERFACE
!
!     Set implicit parameters
!
      if ( present(alpha0) ) then
         alpha = alpha0
      else
         alpha = 1.0D0
      end if
      if ( present(side0) ) then
         side = side0
      else
         side = 'L'
      end if
      if ( present(uplo0) ) then
         uplo = uplo0
      else
         uplo = 'U'
      end if
      if ( present(trans0) ) then
         transA = trans0
      else
         transA = 'N'
      end if
      if ( present(diag0) ) then
         diag = diag0
      else
         diag = 'N'
      end if
!
!     Fetch matrix information
!
      call SLhpf_ddescset3(A, descA, B, descB, P0=P, Q0=Q, info0=info)
      call SLhpf_DescQuery(descA, mA, nA, mbA, nbA, rsrcA, csrcA)
      call SLhpf_DescQuery(descB, mB, nB, mbB, nbB, rsrcB, csrcB)
!
!     If info shows error in descriptor, shift it to correct number
!
      if (info .lt. 0) then
         if (info .lt. -200) then
            info = -1500 - MOD(-info, 100)
         else
            info = -1100 - MOD(-info, 100)
         end if
      end if
!
!     Check for illegal size subsections (fatal error) and alignment
!     restrictions in the present code (forced redistribution)
!
      M = size(B,1)
      N = size(B,2)
      K = size(A,1)
      if (K .ne. size(A,2)) then
         print*,'A must be square for call to PDTRSM!'
         info = -8
         call SLhpf_abort(info)
      end if
      if (mbA .ne. nbA) then
         info = -11000 - NB_
         descA(NB_,:) = descA(MB_,:)
      end if
      if ( (side.eq.'l') .or. (side.eq.'L') ) then
         if (K .ne. M) then
            print*,'PDTRSM: size(A,1) .ne. size(B,1) !!'
            info = -5
            call SLhpf_abort(info)
         end if
         if (rsrcA .ne. rsrcB) then
            info = -15000 - RSRC_
            descB(RSRC_,:) = descA(RSRC_,:)
         end if
         if (nbA .ne. mbB) then
            info = -15000 - MB_
            descB(MB_,:) = descA(NB_,:)
         end if
      else
         if (K .ne. N) then
            print*,'PDTRSM: size(A,1) .ne. size(B,2) !!'
            info = -6
            call SLhpf_abort(info)
         end if
         if (csrcA .ne. csrcB) then
            info = -15000 - CSRC_
            descB(CSRC_,:) = descA(CSRC_,:)
         end if
         if (mbA .ne. nbB) then
            info = -15000 - NB_
            descB(NB_,:) = descA(MB_,:)
         end if
      end if
!
      if (info .eq. 0) then
!
!     No redistribution needed !!
!

         call SLhpf_dtrsm1(side, uplo, transA, diag, m, n, alpha, A,    &
     &                     1, 1, descA, B, 1, 1, descB)
!
      else
!
!     Redistribution needed !!
!
         if (REDIST) then
            write(*,*) "WARNING:  redistribution is necessary"
         end if
!
         info = 0
!
!        Since we are redistributing, use optimal parameters all around
!
         call SLhpf_get_gridshape(P, Q)
         descA(MB_,:) = opt_NB
         descA(NB_,:) = opt_NB
         descA(RSRC_,:) = 0
         descA(CSRC_,:) = 0
         descB(MB_,:) = opt_NB
         descB(NB_,:) = opt_NB
         descB(RSRC_,:) = 0
         descB(CSRC_,:) = 0
         call SLhpf_dtrsm2(P, Q, opt_NB, opt_NB, opt_NB, opt_NB,        &
     &                        side, uplo, transA, diag, m, n, alpha,    &
     &                        A, 1, 1, descA, B, 1, 1, descB)
      endif

      return
      end



