      subroutine SLhpf_sposv(A, B, uplo0, info0)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     September 1, 1998
!     University of Tennessee, Knoxville
!     Code contributors: R. Clint Whaley
!     Last modified by: C. A. Papadopoulos
!
      implicit none
      include 'pblas.h'
      include 'misc.h'
      real(SP), intent(inout), dimension(:,:) :: A, B
      character(LEN=1), optional, intent(in) :: uplo0
      integer, optional, intent(out) :: info0
!HPF$ INHERIT A, B

      character(len=1) :: uplo
      integer :: P, Q, csrcA, csrcB, i, info, j, mA, mB, mbA, mbB, nA,  &
     &           nB, nbA, nbB, rsrcA, rsrcB
      integer, dimension(DLEN_,NUMBER_OF_PROCESSORS()) :: descA, descB
!HPF$ DISTRIBUTE (*,block) :: descA, descB

      INTERFACE

      subroutine SLhpf_sdescset3(A, descA, B, descB, C, descC, P0, Q0,  &
     &                           info0)
         include 'pblas.h'
         include 'misc.h'
         real(SP), dimension(:,:), intent(in) :: A
         integer, dimension(DLEN_,NUMBER_OF_PROCESSORS()),              &
     &            intent(out) :: descA
         real(SP), dimension(:,:), optional, intent(in) :: B, C
         integer, optional, dimension(DLEN_,NUMBER_OF_PROCESSORS()),    &
     &            intent(out) :: descB, descC
         integer, optional, intent(out) :: P0, Q0, info0
!HPF$    INHERIT A, B, C
!HPF$    DISTRIBUTE(*,block) :: descA, descB, descC
      end subroutine SLhpf_sdescset3

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_DescQuery(desc, M, N, MB, NB, RSRC, CSRC)
         include 'pblas.h'
         integer, intent(in) :: desc(:,:)
!HPF$    DISTRIBUTE(*,block) :: desc
         integer, optional, intent(out) :: M, N, MB, NB, RSRC, CSRC
      end subroutine SLhpf_DescQuery

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_sposv1(uplo, n, nrhs, A, ia, ja, descA, B,       &
     &                        ib, jb, descB, info)
         include 'pblas.h'
         include 'misc.h'
         character(LEN=1), intent(in) :: uplo
         integer, intent(in) :: n, nrhs, ia, ja, ib, jb
         integer, dimension(:,:), intent(inout) :: descA, descB
         integer, intent(out) :: info
         real(SP), intent(inout), dimension(:,:) :: A, B
!HPF$    INHERIT A, B
!HPF$    DISTRIBUTE(*,block) :: descA, descB
      end subroutine SLhpf_sposv1

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_sposv2(P, Q, mbA, nbA, mbB, nbB, uplo, n, nrhs,  &
     &                        A, ia, ja, descA, B, ib, jb, descB, info)
         include 'pblas.h'
         include 'misc.h'
         character(LEN=1), intent(in) :: uplo
         integer, intent(in) :: P, Q, mbA, nbA, mbB, nbB, n, nrhs,      &
     &                          ia, ja, ib, jb
         integer, dimension(:,:), intent(inout) :: descA, descB
         integer, intent(out) :: info
         real(SP), intent(inout), dimension(:,:) :: A, B
!HPF$    PROCESSORS PROC(P,Q)
!HPF$ DISTRIBUTE A(CYCLIC(mbA), CYCLIC(nbA)) ONTO PROC
!HPF$ DISTRIBUTE B(CYCLIC(mbB), CYCLIC(nbB)) ONTO PROC
!HPF$ DISTRIBUTE(*,block) :: descA, descB
      end subroutine SLhpf_sposv2

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_get_gridshape(P, Q, maxP, maxQ)
         integer, intent(out) :: P, Q
         integer, optional, intent(in) :: maxP, maxQ
      end subroutine SLhpf_get_gridshape

      extrinsic(hpf_local) subroutine SLhpf_abort(errno)
         integer, intent(in) :: errno
      end subroutine SLhpf_abort

      END INTERFACE

      info = 0

      call SLhpf_sdescset3(A, descA, B, descB, P0=P, Q0=Q, info0=info)
      call SLhpf_DescQuery(descA, mA, nA, mbA, nbA, rsrcA, csrcA)
      call SLhpf_DescQuery(descB, mB, nB, mbB, nbB, rsrcB, csrcB)
!
!     If info shows error in descriptor, shift it to correct number
!
      if (info .lt. 0) then
         if (info .lt. -200) then
            info = -1100 - MOD(-info, 100)
         else
            info = -600 - MOD(-info, 100)
         end if
      end if
!
!     Check for illegal size subsection
!
      if ( (mA .ne. nA) .or. (mA .ne. mB) ) then
         if ( present(info0) ) then
            info0 = -1
         else
            print*,'PSPOSV exited with error code ',info
            call SLhpf_abort(info)
            stop
         end if
         return
      end if
      if (mA .eq. 0) return
!
!     These represent alignment restrictions in present version psposv
!
      if (mbA .ne. nbA) then          ! A must have square blocks
         descA(MB_,:) = opt_NB
         descA(NB_,:) = opt_NB
         if (info .eq. 0) info = -600 - NB_
      end if
      if (mbA .ne. mbB) then          ! force same row blocking
         descB(MB_,:) = descA(MB_,:)
         if (info .eq. 0) info = -1100 - MB_
      end if
      if (rsrcA .ne. rsrcB) then      ! A and B must be row-aligned
         descB(RSRC_,:) = descA(RSRC_,:)
         if (info .eq. 0) info = -1100 - RSRC_
      end if
      if ( present(uplo0) ) then
         uplo = uplo0
      else
         uplo = 'U'
      end if
!
      if (info .eq. 0) then
!
!     No redistribution needed !!
!
         call SLhpf_sposv1(uplo, mA, nB, A, 1, 1, descA, B, 1, 1,       &
     &                     descB, info)
      else
!
!     Redistribution needed !!
!
         if (REDIST) then
            write(*,*) "WARNING:  redistribution is necessary"
         end if
!
         call SLhpf_DescQuery(descA, mA, nA, mbA, nbA)
         call SLhpf_DescQuery(descB, mB, nB, mbB, nbB)
         info = 0
         call SLhpf_get_gridshape(P, Q)
         call SLhpf_sposv2(P, Q, mbA, nbA, mbB, nbB, uplo, mA, nB,      &
     &                        A, 1, 1, descA, B, 1, 1, descB, info)
      end if

      if ( present(info0) ) then
         info0 = info
      else if (info .ne. 0) then
         print*,'PSPOSV exited with error code ',info
         call SLhpf_abort(info)
         stop
      end if

      return
      end


