      subroutine SLhpf_ssyev(A, W, Z, uplo0, info0)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     Written by Caroline A. Papadopoulos, University of Tennessee, Knoxville
!     Adapted from SLhpf routines written by R. Clint Whaley,
!     University of Tennessee, Knoxville
!     September 1, 1998
!
      implicit none
      include 'pblas.h'
      include 'misc.h'

!declaration of parameters
      real(SP), intent(inout), dimension(:,:) :: A
      real(SP), intent(out), dimension(:) :: W
      real(SP), optional, intent(out), dimension(:,:) :: Z
      character(LEN=1), optional, intent(in) :: uplo0
      integer, optional, intent(out) :: info0
!HPF$ INHERIT A, Z
! the following should replicate w accross all processors
!HPF$ ALIGN W(*) with A(*,*)
      integer :: P, Q, csrcA, csrcZ, i, info, j, mA, mZ, mbA, mbZ, nA,  &
     &           nW, nZ, nbA, nbZ, rsrcA, rsrcZ
      integer, dimension(DLEN_,NUMBER_OF_PROCESSORS()) :: descA, descZ
         character(LEN=1) :: jobZ, uplo
!HPF$ DISTRIBUTE (*,block) :: descA, descZ

      INTERFACE

      subroutine SLhpf_sdescset3(A, descA, B, descB, C, descC, P0, Q0,  &
     &                           info0)
         include 'pblas.h'
         include 'misc.h'
         real(SP), dimension(:,:), intent(in) :: A
         integer, dimension(DLEN_,NUMBER_OF_PROCESSORS()),              &
     &            intent(out) :: descA
         real(SP), dimension(:,:), optional, intent(in) :: B, C
         integer, optional, dimension(DLEN_,NUMBER_OF_PROCESSORS()),    &
     &            intent(out) :: descB, descC
         integer, optional, intent(out) :: P0, Q0, info0
!HPF$    INHERIT A, B, C
!HPF$    DISTRIBUTE(*,block) :: descA, descB, descC
      end subroutine SLhpf_sdescset3

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_DescQuery(desc, M, N, MB, NB, RSRC, CSRC)
         include 'pblas.h'
         integer, intent(in) :: desc(:,:)
!HPF$    DISTRIBUTE(*,block) :: desc
         integer, optional, intent(out) :: M, N, MB, NB, RSRC, CSRC
      end subroutine SLhpf_DescQuery

      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_ssyev1(jobz, uplo, n, A, ia, ja, descA, W,       &
     &                        Z, iz, jz, descZ, info)
      implicit none
      include 'pblas.h'
      include 'misc.h'

      character(LEN=1), intent(in) :: jobz, uplo
      integer, intent(in) :: n, ia, ja, iz, jz
      integer, intent(inout), dimension(:,:) :: descA, descZ
      integer, intent(out) :: info
      real(SP), intent(inout), dimension(:,:) :: A
      real(SP), optional, intent(out), dimension(:,:) :: Z
      real(SP), intent(out), dimension(:) :: W
!HPF$ INHERIT A, Z
!HPF$ ALIGN W(*) with A(*,*)
!HPF$ DISTRIBUTE(*,block) :: descA, descZ
      end subroutine SLhpf_ssyev1


      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_ssyev2(P, Q, jobz, uplo, n, A, ia, ja,descA,     &
     &                        W, Z, iz, jz, descZ, info, mbA, nbA,      &
     &                        mbZ, nbZ)
      include 'pblas.h'
      include 'misc.h'
      character(LEN=1), intent(in) :: jobz, uplo
      integer, intent(in) :: P, Q, n, ia, ja, iz, jz, mbA,              &
     &                       nbA, mbZ, nbZ
      integer, dimension(:,:), intent(inout) :: descA, descZ
      integer, intent(out) :: info
      real(SP), intent(inout), dimension(:,:) :: A
      real(SP), optional, intent(out), dimension(:,:) :: Z
      real(SP), intent(out), dimension(:) :: W
!HPF$ PROCESSORS PROC(P,Q)
!HPF$ DISTRIBUTE A(CYCLIC(mbA), CYCLIC(nbA)) ONTO PROC
!HPF$ DISTRIBUTE Z(CYCLIC(mbZ), CYCLIC(nbZ)) ONTO PROC
!HPF$ ALIGN W(*)  WITH A(*,*)
!HPF$ DISTRIBUTE(*,block) :: descA, descZ
      end subroutine SLhpf_ssyev2


      extrinsic(hpf_local)                                              &
     &subroutine SLhpf_get_gridshape(P, Q, maxP, maxQ)
         integer, intent(out) :: P, Q
         integer, optional, intent(in) :: maxP, maxQ
      end subroutine SLhpf_get_gridshape

      extrinsic(hpf_local) subroutine SLhpf_abort(errno)
         integer, intent(in) :: errno
      end subroutine SLhpf_abort

      END INTERFACE

! find out the sizes fo the matricies A, Z and W and update
! the descriptors

      info = 0
      nW   = size(W)

      if (present(Z)) then
         call SLhpf_sdescset3(A, descA, Z, descZ, P0=P, Q0=Q,           &
     &                             info0=info)
         call SLhpf_DescQuery(descA, mA, nA, mbA, nbA, rsrcA, csrcA)
         call SLhpf_DescQuery(descZ, mZ, nZ, mbZ, nbZ, rsrcZ, csrcZ)
      else
         call SLhpf_sdescset3(A, descA, P0=P, Q0=Q, info0=info)
         call SLhpf_DescQuery(descA, mA, nA, mbA, nbA, rsrcA, csrcA)
      endif
!
!     If info shows error in descriptor, shift it to correct number
!
      if (info .lt. 0) then
         if (info .lt. -200) then
            info = -1200 - MOD(-info, 100)
         else
            info = -700 - MOD(-info, 100)
         end if
      end if
!
!     Check for illegal size subsection
!
! A must be square nxn
      if  (mA.ne.nA)  then
         info = -4
         if ( present(info0) ) then
            info0 = info
            return
         else
            print*,'PSSYEV exited with error code ',info
            call SLhpf_abort(info)
            stop
         end if
!
! W must have length n
!
      elseif  (mA.ne.nW)  then
         info = -8
         if ( present(info0) ) then
            info0 = info
            return
         else
            print*,'PSSYEV exited with error code ',info
            call SLhpf_abort(info)
            stop
         end if
!
! Z must be square nxn
!
      elseif ( present(Z) ) then
         if  (mA.ne.mZ)  then
            info = -1200 - M_
            if ( present(info0) ) then
               info0 = info
               return
            else
               print*,'PSSYEV exited with error code ',info
               call SLhpf_abort(info)
               stop
            end if
         end if
         if  (nA.ne.nZ)  then
            info = -1200 - N_
            if ( present(info0) ) then
               info0 = info
               return
            else
               print*,'PSSYEV exited with error code ',info
               call SLhpf_abort(info)
               stop
            end if
         end if
      end if
      if (mA .eq. 0) return
!
!     These represent alignment restrictions in present version pssyev
!
      if (mbA .ne. nbA) then          ! A must have square blocks
         descA(MB_,:) = opt_NB
         descA(NB_,:) = opt_NB
         if (info .eq. 0) info = -700 - NB_
      end if

      if (present(Z)) then
         if (mbA .ne. mbZ) then          ! force same row blocking
            descZ(MB_,:) = descA(MB_,:)
            if (info .eq. 0) info = -1200 - MB_
         end if
         if (nbA .ne. nbZ) then          ! force same column blocking
            descZ(NB_,:) = descA(NB_,:)
            if (info .eq. 0) info = -1200 - NB_
         end if
         if (rsrcA .ne. rsrcZ) then      ! A and Z must be row-aligned
            descZ(RSRC_,:) = descA(RSRC_,:)
            if (info .eq. 0) info = -1200 - RSRC_
         end if
      end if
!
! initialize uplo, jobz, and descZ to call layer 2
!
      if ( present(uplo0) ) then
         uplo = uplo0
      else
         uplo = 'U'
      end if

      if ( present(Z) ) then
         jobZ = 'V'
      else
         jobZ = 'N'
         descZ = descA
      end if

      if (info .eq. 0) then
!
!     No redistribution needed!!
!
         if (present(Z)) then
            call SLhpf_ssyev1(jobZ, uplo, mA, A, 1, 1, descA, W, Z,     &
     &                     1, 1, descZ, info)
         else
            call SLhpf_ssyev1(jobZ, uplo, mA, A, 1, 1, descA, W, Z,     &
     &                     iz=1, jz=1, descZ= descZ, info=info)
         end if
!
      else
!
!     Redistribution needed !!
!
         if (REDIST) then
            write(*,*) "WARNING:  redistribution is necessary"
         end if
!
         call SLhpf_DescQuery(descA, mA, nA, mbA, nbA)
         call SLhpf_DescQuery(descZ, mZ, nZ, mbZ, nbZ)
         info = 0
         call SLhpf_get_gridshape(P, Q)
         if (present(Z)) then
            call SLhpf_ssyev2(P, Q, jobZ, uplo, mA, A, 1, 1, descA, W,  &
     &                        Z, 1, 1, descZ, info, mbA, nbA, mbZ, nbZ)
         else
            call SLhpf_ssyev2(P, Q, jobZ, uplo, mA, A, 1, 1, descA, W,  &
     &                        iz=1, jz=1, descZ=descZ, info=info,       &
     &      mbA = mbA, nbA = nbA, mbZ = mbZ, nbZ = nbZ)
         end if
      end if

      if ( present(info0) ) then
         info0 = info
      else if (info .ne. 0) then
         print*,'PSSYEV exited with error code ',info
         call SLhpf_abort(info)
         stop
      end if

      return
      end


