      subroutine SLhpf_zmatinf(A, descA, info0)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     September 1, 1998
!     University of Tennessee, Knoxville
!     Code contributors: R. Clint Whaley
!     Last modified by: C. A. Papadopoulos
!
      use hpf_library
      implicit none
      include 'misc.h'
      include 'pblas.h'
      complex(DP), intent(in) :: A(:,:)
      integer, dimension(DLEN_,NUMBER_OF_PROCESSORS()),                 &
     &         intent(out) :: descA
!HPF$ DISTRIBUTE (*,block) :: descA
      integer, optional, intent(out) :: info0
!HPF$ INHERIT A

      character(len=9), dimension(2) :: axis_type
      integer :: axis_info(2)
      integer :: m, n, mb, nb, info
      logical :: IDENTITY_MAP
      integer :: tmp(2), LB(2), STRIDE(2), AXIS_MAP(2)

      INTERFACE
      logical function SLhpf_strcmp(str1, str2)
         character (*), intent(in) :: str1, str2
      end function SLhpf_strcmp
      END INTERFACE

      info = 0
!
!     Figure M_A and N_A
!
      M = size(A, 1)
      N = size(A, 2)
!
!     Figure MB_A and NB_A
!
      call HPF_DISTRIBUTION(A, AXIS_TYPE, AXIS_INFO)

      if (SLhpf_strcmp(AXIS_TYPE(1)(1:5), 'BLOCK')) then
         mb = AXIS_INFO(1)
      else if (SLhpf_strcmp(AXIS_TYPE(1)(1:9), 'COLLAPSED')) then
         mb = M
      else if (SLhpf_strcmp(AXIS_TYPE(1)(1:6), 'CYCLIC')) then
         mb = AXIS_INFO(1)
      else
         info = -MB_
         mb = OPT_NB
      end if

      if (SLhpf_strcmp(AXIS_TYPE(2)(1:5), 'BLOCK')) then
         nb = AXIS_INFO(2)
      else if (SLhpf_strcmp(AXIS_TYPE(2)(1:9), 'COLLAPSED')) then
         nb = N
      else if (SLhpf_strcmp(AXIS_TYPE(2)(1:6), 'CYCLIC')) then
         nb = AXIS_INFO(2)
      else
         info = -NB_
         nb = OPT_NB
      end if
!
!     If object to which A is ultimately aligned is correctly distributed, make
!     sure A itself is
!
      if (info.eq.0) then
         call HPF_ALIGNMENT(A, LB, tmp, stride, axis_map, identity_map)
!
!        If not ultimately aligned to itself, need to see if other
!        constraints hold true
!
         if (.not.identity_map) then
!
!           For now, require that matrix be aligned to same dimensions of
!           ultimate align target
!
            if (axis_map(1) .ne. 1) then
               info = -M_
            else if (axis_map(2) .ne. 2) then
               info = -N_
            else
!
!              NOTE: we can relax the following stride restrictions if
!                    we take into account information about the process
!                    grid.  For simplicity, and because it would only rarely
!                    help, we do not do this.
!

               if (stride(1) .ne. 1) then
                  if (stride(1) .eq. mb) then
                     mb = 1
                  else if (M .le. mb/stride(1)) then ! axis 1 not distributed
                     mb = M
                  else
                     info = -MB_
                  end if
               end if
!
!              Error on partial first block if axis is distributed
!
               if (mod(LB(1)-1, mb) .ne. 0) then
                  if (M .gt. mb) info = -MB_
               end if
!
!              Do same checks on 2nd axis
!
               if (stride(2) .ne. 1) then
                  if (stride(2) .eq. nb) then
                     nb = 1
                  else if (N .le. nb/stride(2)) then ! axis 2 not distributed
                     nb = N
                  else
                     if (info .eq. 0) info = -NB_
                  end if
               end if
!
!              Error on partial first block if axis is distributed
!
               if (mod(LB(2)-1, nb) .ne. 0) then
                  if (N .gt. nb) then
                     if (info .eq. 0) info = -NB_
                  end if
               end if
            end if
         end if
      end if

!
!     If the matrix requires redistribution, set descriptor to what we
!     believe is optimal
!
      descA(DT_,:) = BLOCK_CYCLIC_2D
      descA(M_,:) = M
      descA(N_,:) = N
      if (info .ne. 0) then
         descA(MB_,:) = OPT_NB
         descA(NB_,:) = OPT_NB
      else
         descA(MB_,:) = mb
         descA(NB_,:) = nb
      end if
      if (present(info0)) info0 = info

      return
      end

