      double precision function dmatnrm(A)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     September 1, 1998
!     Written by R. Clint Whaley, University of Tennessee, Knoxville
!
      include 'misc.h'
      real(DP), intent(in), dimension(:,:) :: A
      integer :: i
      real(DP) :: maxnrm, tmp

      maxnrm = 0.0D0
      do i = 1, size(A,1)
         tmp = sum( abs(A(i,:)) )
         if (tmp .gt. maxnrm) maxnrm = tmp
      end do

      dmatnrm = maxnrm

      RETURN
      END

      double precision function dmatnrm_v(V)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     Written by Caroline Papadopoulos, University of Tennessee, Knoxville
!     Adapted from SLhpf routines written by R. Clint Whaley,
!     University of Tennessee, Knoxville
!     September 1, 1998
!
      include 'misc.h'
      real(DP), intent(in), dimension(:) :: V
      integer :: i
      real(DP) :: maxnrm, tmp

      maxnrm = 0.0D0
      do i = 1, size(V,1)
         tmp = abs(V(i))
         if (tmp .gt. maxnrm) maxnrm = tmp
      end do

      dmatnrm_v = maxnrm

      RETURN
      END

      subroutine posvinfo( thresh, ngrid, nmat, nop, Pval, Qval, mAval, &
     &                     nAval, mbAval, nbAval, rsrcAval, csrcAval,   &
     &                     mBval, nBval, mbBval, nbBval, rsrcBval,      &
     &                     csrcBval, UPLOval, Mval, Nval, IAval, JAval, &
     &                     IBval, JBval )
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     September 1, 1998
!     Written by R. Clint Whaley, University of Tennessee, Knoxville
!
      real, intent(out) :: thresh
      integer, intent(out) :: ngrid, nmat, nop
      integer, pointer, dimension(:) :: IAval, IBval, JAval, JBval,     &
     &         Mval, Nval, Pval, Qval, UPLOval, csrcAval, csrcBval,     &
     &         mAval, mBval, mbAval, mbBval, nAval, nBval, nbAval,      &
     &         nbBval, rsrcAval, rsrcBval
      character(LEN=1), allocatable, dimension(:) :: ctmp
      integer, parameter :: nin=11
      integer :: i

      open(nin, file='posv.dat', STATUS='OLD')

      read(nin,*) ngrid
      allocate( Pval(ngrid), Qval(ngrid) )
      read(nin,*) ( Pval(i), i = 1, ngrid )
      read(nin,*) ( Qval(i), i = 1, ngrid )

      read(nin,*) nmat
      allocate( mAval(nmat), nAval(nmat), mbAval(nmat), nbAval(nmat),   &
     &          rsrcAval(nmat), csrcAval(nmat), mBval(nmat),            &
     &          nBval(nmat), mbBval(nmat), nbBval(nmat), rsrcBval(nmat),&
     &          csrcBval(nmat) )

      read(nin,*) ( mAval(i), i = 1, nmat )
      read(nin,*) ( nAval(i), i = 1, nmat )
      read(nin,*) ( mbAval(i), i = 1, nmat )
      read(nin,*) ( nbAval(i), i = 1, nmat )
      read(nin,*) ( rsrcAval(i), i = 1, nmat )
      read(nin,*) ( csrcAval(i), i = 1, nmat )
      read(nin,*) ( mBval(i), i = 1, nmat )
      read(nin,*) ( nBval(i), i = 1, nmat )
      read(nin,*) ( mbBval(i), i = 1, nmat )
      read(nin,*) ( nbBval(i), i = 1, nmat )
      read(nin,*) ( rsrcBval(i), i = 1, nmat )
      read(nin,*) ( csrcBval(i), i = 1, nmat )

      read(nin,*) nop
      allocate( Mval(nop), Nval(nop), IAval(nop), JAval(nop),           &
     &          IBval(nop), JBval(nop), UPLOval(nop) )
!
!     Store character info in an integer array to get around pghpf compiler
!     error with having character arrays have the pointer attribute
!
      allocate( ctmp(nop) )
      read(nin,*) ( ctmp(i), i = 1, nop )
      do i = 1, nop
         if ( (ctmp(i).eq.'U') .or. (ctmp(i).eq.'u') ) then
            UPLOval(i) = 1
         else
            if ( (ctmp(i).ne.'L') .and. (ctmp(i).ne.'l') ) then
               print*,'Unknown UPLO value ''',ctmp(i),                  &
     &                ''' assumed to be L.'
            end if
            UPLOval(i) = 2
         end if
      end do
      deallocate( ctmp )
      read(nin,*) ( Mval(i), i = 1, nop )
      read(nin,*) ( Nval(i), i = 1, nop )
      read(nin,*) ( IAval(i), i = 1, nop )
      read(nin,*) ( JAval(i), i = 1, nop )
      read(nin,*) ( IBval(i), i = 1, nop )
      read(nin,*) ( JBval(i), i = 1, nop )
      read(nin,*) thresh

      close(nin)

      return
      end subroutine posvinfo

      program dposvtst

!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     September 1, 1998
!     University of Tennessee, Knoxville
!     Code contributors: R. Clint Whaley
!     Last modified by: C. A. Papadopoulos
!
      include 'misc.h'
      real(DP), allocatable, dimension(:,:) :: A, X, B
      real(DP), allocatable, dimension(:) :: X_V, B_V

      INTERFACE
      extrinsic(hpf_local) subroutine SLhpf_exit()
      end subroutine SLhpf_exit

      subroutine dposvtst2(thresh, N0, P, Q, mb_A, nb_A, mb_B, nb_B,    &
     &                     rsrc_A, csrc_A, rsrc_B, csrc_B, Nop, UPLOval,&
     &                     Mval, Nval, IAval, JAval, IBval, JBval, A, B,&
     &                     npass, nfail, nskip, ninfo, nbyp)
         include 'misc.h'
         integer, intent(in) :: N0, P, Q, mb_A, nb_A, mb_B, nb_B,       &
     &                          rsrc_A, csrc_A, rsrc_B, csrc_B, Nop
         integer, intent(in), dimension(:) :: IAval, IBval, JAval,      &
     &                                        JBval, Mval, Nval,        &
     &                                        UPLOval
         integer, intent(out) :: npass, nfail, nskip, ninfo, nbyp
         real, intent(in) :: thresh
         real(DP), dimension(:,:) :: A, B
!HPF$ PROCESSORS PROC(P,Q)
!HPF$ TEMPLATE, DIMENSION(N0, N0) :: ATEMP, BTEMP
!HPF$ DISTRIBUTE ATEMP(cyclic(mb_A), cyclic(nb_A)) ONTO PROC
!HPF$ DISTRIBUTE BTEMP(cyclic(mb_B), cyclic(nb_B)) ONTO PROC
!HPF$ ALIGN A(I,J) WITH ATEMP(rsrc_A*mb_A+I, csrc_A*nb_A+J)
!HPF$ ALIGN B(I,J) WITH BTEMP(rsrc_B*mb_B+I, csrc_B*nb_B+J)
      end subroutine dposvtst2


      subroutine dposvtst2_v(thresh, N0, P, Q, mb_A, nb_A, mb_B, rsrc_A,&
     &                       csrc_A, rsrc_B, csrc_B, Nop, UPLOval, Mval,&
     &                       Nval, IAval, JAval, IBval, A, B_V, npass,  &
     &                       nfail, nskip, ninfo, nbyp)
         include 'misc.h'
         integer, intent(in) :: N0, P, Q, mb_A, nb_A, mb_B,             &
     &                          rsrc_A, csrc_A, rsrc_B, csrc_B, Nop
         integer, intent(in), dimension(:) :: IAval, IBval, JAval, Mval,&
     &                                        Nval, UPLOval
         integer, intent(out) :: npass, nfail, nskip, ninfo, nbyp
         real, intent(in) :: thresh
         real(DP), dimension(:,:) :: A
         real(DP), dimension(:) :: B_V
!HPF$ PROCESSORS PROC(P,Q)
!HPF$ TEMPLATE, DIMENSION(N0, N0) :: ATEMP
!HPF$ DISTRIBUTE ATEMP(cyclic(mb_A), cyclic(nb_A)) ONTO PROC
!HPF$ ALIGN A(I,J) WITH ATEMP(rsrc_A*mb_A+I, csrc_A*nb_A+J)
!HPF$ ALIGN B_V(I) WITH ATEMP(rsrc_B*mb_B+I,1)
      end subroutine dposvtst2_v


      subroutine posvinfo(thresh, ngrid, nmat, nop, Pval, Qval, mAval,  &
     &                    nAval, mbAval, nbAval, rsrcAval, csrcAval,    &
     &                    mBval, nBval, mbBval, nbBval, rsrcBval,       &
     &                    csrcBval, UPLOval, Mval, Nval, IAval, JAval,  &
     &                    IBval, JBval)
         real, intent(out) :: thresh
         integer, intent(out) :: ngrid, nmat, nop
         integer, pointer, dimension(:) :: IAval, IBval, JAval, JBval,  &
     &                                     Mval, Nval, Pval, Qval,      &
     &                                     UPLOval, csrcAval, csrcBval, &
     &                                     mAval, mBval, mbAval, mbBval,&
     &                                     nAval, nBval, nbAval, nbBval,&
     &                                     rsrcAval, rsrcBval
      end subroutine posvinfo
      END INTERFACE

      integer :: igrid, imat, itest, k, nbyp, nbypa, nfai, nfail, ngrid,&
     &           ninf, ninfo, nmat, nop, npas, npass, nski, nskip,      &
     &           ntest
      integer, pointer, dimension(:) :: IAval, IBval, JAval, JBval,     &
     &                                  M_Aval, M_Bval, Mval, N_Aval,   &
     &                                  N_Bval, Nval, Pval, Qval,       &
     &                                  UPLOval, csrc_Aval, csrc_Bval,  &
     &                                  mb_Aval, mb_Bval, nb_Aval,      &
     &                                  nb_Bval, rsrc_Aval, rsrc_Bval
      real :: thresh

      call posvinfo(thresh, ngrid, nmat, nop, Pval, Qval, M_Aval,       &
     &              N_Aval, mb_Aval, nb_Aval, rsrc_Aval, csrc_Aval,     &
     &              M_Bval, N_Bval, mb_Bval, nb_Bval, rsrc_Bval,        &
     &              csrc_Bval, UPLOval, Mval, Nval, IAval, JAval, IBval,&
     &              JBval)
      itest = 1
      nfail = 0; npass = 0; nskip = 0; ninfo = 0; ntest = 0; nbypa = 0
!
!     Find upper bound for template size
!
      k = max( maxval(MB_Aval), maxval(NB_Aval), maxval(MB_Bval),       &
     &         maxval(NB_Bval) )
      k = k * max( maxval(rsrc_Aval), maxval(csrc_Aval),                &
     &             maxval(rsrc_Bval), maxval(csrc_Bval) )
      k = k + max( maxval(M_Aval), maxval(N_Aval), maxval(M_Bval),      &
     &             maxval(N_Bval) )
      do igrid = 1, ngrid
         do imat = 1, nmat
            write(*,*) ' '
            write(*,5000)
            write(*,1000) itest
            write(*,2000) Pval(igrid), Qval(igrid)
            write(*,3000) M_Aval(imat), N_Aval(imat), mb_Aval(imat),    &
     &                    nb_Aval(imat), rsrc_Aval(imat),               &
     &                    csrc_Aval(imat)
            write(*,4000) M_Bval(imat), N_Bval(imat), mb_Bval(imat),    &
     &                    nb_Bval(imat), rsrc_Bval(imat),               &
     &                    csrc_Bval(imat)
            write(*,5000)
            allocate( A(M_Aval(imat), N_Aval(imat)) )
            allocate( B(M_Bval(imat), N_Bval(imat)) )
            allocate( B_V(M_Bval(imat)) )
            A = -99.9D10
            B = -99.8D10
            B_V = -99.8D10
            call dposvtst2(thresh, k, Pval(igrid), Qval(igrid),         &
     &                     mb_Aval(imat), nb_Aval(imat), mb_Bval(imat), &
     &                     nb_Bval(imat), rsrc_Aval(imat),              &
     &                     csrc_Aval(imat), rsrc_Bval(imat),            &
     &                     csrc_Bval(imat), Nop, UPLOval, Mval, Nval,   &
     &                     IAval, JAval, IBval, JBval, A, B, npas, nfai,&
     &                     nski, ninf, nbyp)
            npass = npass + npas
            nfail = nfail + nfai
            nskip = nskip + nski
            ninfo = ninfo + ninf
            nbypa = nbypa + nbyp
            write(*,5000)
            write(*,6000) itest
            write(*,5000)
            write(*,*) ' '
            call dposvtst2_v(thresh, k, Pval(igrid), Qval(igrid),       &
     &                       mb_Aval(imat), nb_Aval(imat),              &
     &                       mb_Bval(imat), rsrc_Aval(imat),            &
     &                       csrc_Aval(imat), rsrc_Bval(imat),          &
     &                       csrc_Bval(imat), Nop, UPLOval, Mval, Nval, &
     &                       IAval, JAval, IBval, A, B_V, npas, nfai,   &
     &                       nski, ninf, nbyp)
            npass = npass + npas
            nfail = nfail + nfai
            nskip = nskip + nski
            ninfo = ninfo + ninf
            nbypa = nbypa + nbyp
            write(*,5000)
            write(*,6000) itest
            write(*,5000)
            write(*,*) ' '
            deallocate(B_V)
            deallocate(A)
            deallocate(B)
            itest = itest + 1
         end do
      end do
      deallocate( Pval, Qval, m_Aval, n_Aval, mb_Aval, nb_Aval,         &
     &            rsrc_Aval, csrc_Aval, m_Bval, n_Bval, mb_Bval,        &
     &            nb_Bval, rsrc_Bval, csrc_Bval, UPLOval, Mval, Nval )

      write(*,*) ' '
      write(*,*) ' '
      write(*,*) ' '
      write(*,5000)
      write(*,5000)
      write(*,7000) 2*Nop*(itest-1), npass, nfail, nskip, ninfo, nbypa
      write(*,5000)
      write(*,5000)
      call SLhpf_exit()

 1000 format(x,'Begin test batch #',i6,':')
 2000 format(x,'P =',i4, '; Q =',i4,';')
 3000 format(x,'M_A =',i6,'; N_A =',i6,'; MB_A =',i4,'; NB_A =',i4,     &
     &       '; RSRC_A =',i3,'; CSRC_A =',i3)
 4000 format(x,'M_B =',i6,'; N_B =',i6,'; MB_B =',i4,'; NB_B =',i4,     &
     &       '; RSRC_B =',i3,'; CSRC_B =',i3)
 5000 format(x,'******************************************************' &
     &         '*************************')
 6000 format(x,'Done  test batch #',i6,'.')
 7000 format(x,I5,' TESTS:',I5,' PASSED,',I5,' FAILED,',I5,' SKIPPED,', &
     &       I5,' BAD INFO',I5,' BYPASS')

      stop
      end

      subroutine dposvtst2(thresh, N0, P, Q, mb_A, nb_A, mb_B, nb_B,    &
     &                     rsrc_A, csrc_A, rsrc_B, csrc_B,              &
     &                     Nop, UPLOval, Mval, Nval, IAval, JAval,      &
     &                     IBval, JBval, A, B, npass, nfail, nskip,     &
     &                     ninfo, nbyp)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     September 1, 1998
!     Written by R. Clint Whaley, University of Tennessee, Knoxville
!
      USE HPF_LAPACK
      include 'misc.h'
      character(LEN=1) :: uplo
      integer, intent(in), dimension(:) :: UPLOval
      integer, intent(in) :: N0, P, Q, mb_A, nb_A, mb_B, nb_B,          &
     &                       rsrc_A, csrc_A, rsrc_B, csrc_B, Nop
      integer, intent(in), dimension(:) :: Mval, Nval, IAval, JAval,    &
     &                                     IBval, JBval
      integer, intent(out) :: npass, nfail, nskip, ninfo, nbyp
      real, intent(in) :: thresh
      real(DP), dimension(:,:) :: A, B
!HPF$ PROCESSORS PROC(P,Q)
!HPF$ TEMPLATE, DIMENSION(N0, N0) :: ATEMP, BTEMP
!HPF$ DISTRIBUTE ATEMP(cyclic(mb_A), cyclic(nb_A)) ONTO PROC
!HPF$ DISTRIBUTE BTEMP(cyclic(mb_B), cyclic(nb_B)) ONTO PROC
!HPF$ ALIGN A(I,J) WITH ATEMP(rsrc_A*mb_A+I, csrc_A*nb_A+J)
!HPF$ ALIGN B(I,J) WITH BTEMP(rsrc_B*mb_B+I, csrc_B*nb_B+J)

      character*7 stat
      real(DP), allocatable, dimension(:,:) :: A0, X
      real(DP) normA, eps, sres, fres
      integer :: i, j, k, ia, ja, ib, jb, info
      integer :: N, AM, AN, BM, BN, NRHS,                               &
     &           IAbeg, IAend, IAinc, JAbeg, JAend, JAinc,              &
     &           IBbeg, IBend, IBinc, JBbeg, JBend, JBinc
      real(DP), dimension(NUMBER_OF_PROCESSORS()) :: start, times

      INTERFACE
      double precision function dpocheck(A, LU, uplo, normA0, eps0)
      include 'misc.h'
      real(DP), dimension(:,:), intent(in) :: A, LU
      character(LEN=1), intent(in) :: uplo
      real(DP), optional, intent(in) :: normA0, eps0
      end function dpocheck

      double precision function dslvcheck(A, x, b, normA0, eps0)
      include 'misc.h'
      real(DP), dimension(:,:), intent(in) :: A, x, b
      real(DP), optional, intent(out) :: normA0, eps0
      end function dslvcheck

      extrinsic(hpf_local) function SLhpf_walltime()
         include 'misc.h'
         include 'pblas.h'
         real(DP), dimension(NUMBER_OF_PROCESSORS()) :: SLhpf_walltime
!HPF$    DISTRIBUTE(block) :: SLhpf_walltime
      end function SLhpf_walltime
      END INTERFACE

      print*,'STATUS  UP     N  NRHS    IA    JA    IB    JB  INFO'//   &
     &       '    SOLVE     FACT WALLTIME'
      print*,'======= == ===== ===== ===== ===== ===== ===== ====='//   &
     &       ' ======== ======== ========'

      npass = 0; nfail = 0; nskip = 0; ninfo = 0; nbyp = 0
      do k = 1, Nop
         if (UPLOval(k) .eq. 1) then
            uplo = 'U'
         else
            uplo = 'L'
         end if
         N = Mval(k)
         NRHS = Nval(k)
         IA = IAval(k)
         JA = JAval(k)
         IB = IBval(k)
         JB = JBval(k)
         write(*,1000) 'RUNNING', uplo, N, NRHS, IA, JA, IB, JB, 0,     &
     &                 0.0D0, 0.0D0, 0.0D0
         AM = N + IA-1
         AN = N + JA-1
         BM = N + IB-1
         BN = NRHS + JB-1
         if ( (AM.le.size(A,1)) .and. (AN.le.size(A,2)) .and.           &
     &        (BM.le.size(B,1)) .and. (BN.le.size(B,2)) ) then
            call random_number(A(IA:AM, JA:AN))
            call random_number(B(IB:BM, JB:BN))
!
!           For ease of testing, actually reflect matrix, and force it to
!           be diagonally dominant
!
            forall(i=0:N-1, j=0:N-1, i.lt.j) A(IA+i,JA+j) = A(IA+j,JA+i)
            forall (i=0:N-1) A(IA+i,JA+i) = sum(abs(A(IA:AM,JA+i)))


            if (thresh .ge. 0.0E0) then
               allocate( A0(N, N), X(N, NRHS) )
               X = B(IB:BM, JB:BN)
               A0 = A(IA:AM, JA:AN)
            end if

            start = SLhpf_walltime()
            call la_posv(A(IA:AM, JA:AN), B(IB:BM, JB:BN), uplo, info)
            times = SLhpf_walltime() - start

            sres = -1.0D0
            fres = -1.0D0
            if (info .eq. 0) then
               if (thresh .ge. 0.0E0) then
                  sres = dslvcheck(A0, B(IB:BM, JB:BN), X, normA, eps)
                  if (sres .gt. thresh) then
                     stat = 'FAILED '
                     fres = dpocheck(A0, A(IA:AM, JA:AN), uplo, normA,  &
     &                               eps)
                     nfail = nfail + 1
                  else
                     stat = 'PASSED '
                     npass = npass + 1
                  end if
                  deallocate( A0, X )
               else
                  nbyp = nbyp + 1
                  stat = 'BYPASS '
                  sres = 0.0D0
                  fres = 0.0D0
               end if
            else
               ninfo = ninfo + 1
               stat = 'BADINFO'
               sres = dble( info )
               fres = dble( info )
            end if
         else
            stat = 'SKIPPED'
            nskip = nskip + 1
         end if
         write(*,1000) stat, uplo, N, NRHS, IA, JA, IB, JB, INFO, sres, &
     &                 fres, maxval(times)
      end do
      write(*,2000) Nop, npass, nfail, nskip, ninfo, nbyp

 1000 format(x, a7, 2x, a1, 7I6, 2E9.1, F9.1)
 2000 format(x,I5,' TESTS:',I5,' PASSED,',I5,' FAILED,',I5,' SKIPPED,', &
     &       I5,' BAD INFO', I5, ' BYPASS.')

      RETURN
      END


      subroutine dposvtst2_v(thresh, N0, P, Q, mb_A, nb_A, mb_B,        &
     &                     rsrc_A, csrc_A, rsrc_B, csrc_B,              &
     &                     Nop, UPLOval, Mval, Nval, IAval, JAval,      &
     &                     IBval, A, B_V, npass, nfail, nskip,          &
     &                     ninfo, nbyp)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     Written by C. A. Papadopoulos, University of Tennessee, Knoxville
!     Adapted from SLhpf routines written by R. Clint Whaley,
!     University of Tennessee, Knoxville
!     September 1, 1998
!
      USE HPF_LAPACK
      include 'misc.h'
      character(LEN=1) :: uplo
      integer, intent(in), dimension(:) :: UPLOval
      integer, intent(in) :: N0, P, Q, mb_A, nb_A, mb_B,                &
     &                       rsrc_A, csrc_A, rsrc_B, csrc_B, Nop
      integer, intent(in), dimension(:) :: Mval, Nval, IAval, JAval,    &
     &                                     IBval
      integer, intent(out) :: npass, nfail, nskip, ninfo, nbyp
      real, intent(in) :: thresh
      real(DP), dimension(:,:) :: A
      real(DP), dimension(:) :: B_V
!HPF$ PROCESSORS PROC(P,Q)
!HPF$ TEMPLATE, DIMENSION(N0, N0) :: ATEMP
!HPF$ DISTRIBUTE ATEMP(cyclic(mb_A), cyclic(nb_A)) ONTO PROC
!HPF$ ALIGN A(I,J) WITH ATEMP(rsrc_A*mb_A+I, csrc_A*nb_A+J)
!HPF$ ALIGN B_V(I) WITH ATEMP(rsrc_B*mb_B+I,1)

      character*7 stat
      real(DP), allocatable, dimension(:,:) :: A0
      real(DP), allocatable, dimension(:) :: X_V
      real(DP) normA, eps, sres, fres
      integer :: i, j, k, ia, ja, ib, jb, info
      integer :: N, AM, AN, BM, BN, NRHS,                               &
     &           IAbeg, IAend, IAinc, JAbeg, JAend, JAinc,              &
     &           IBbeg, IBend, IBinc, JBbeg, JBend, JBinc
      real(DP), dimension(NUMBER_OF_PROCESSORS()) :: start, times

      INTERFACE
      double precision function dpocheck(A, LU, uplo, normA0, eps0)
      include 'misc.h'
      real(DP), dimension(:,:), intent(in) :: A, LU
      character(LEN=1), intent(in) :: uplo
      real(DP), optional, intent(in) :: normA0, eps0
      end function dpocheck

      double precision function dslvcheck_v(A, x_v, b_v, normA0, eps0)
      include 'misc.h'
      real(DP), dimension(:,:), intent(in) :: A
      real(DP), dimension(:), intent(in) :: x_v, b_v
      real(DP), optional, intent(out) :: normA0, eps0
      end function dslvcheck_v

      extrinsic(hpf_local) function SLhpf_walltime()
         include 'misc.h'
         include 'pblas.h'
         real(DP), dimension(NUMBER_OF_PROCESSORS()) :: SLhpf_walltime
!HPF$    DISTRIBUTE(block) :: SLhpf_walltime
      end function SLhpf_walltime
      END INTERFACE

      print*,'STATUS  UP   N     IA    JA    IB  INFO'//                &
     &       '    SOLVE     FACT  WALLTIME'
      print*,'======= == ===== ===== ===== ===== ====='//               &
     &       ' ======== ======== ========'

      npass = 0; nfail = 0; nskip = 0; ninfo = 0; nbyp = 0
      do k = 1, Nop
         if (UPLOval(k) .eq. 1) then
            uplo = 'U'
         else
            uplo = 'L'
         end if
         N = Mval(k)
         NRHS = Nval(k)
         IA = IAval(k)
         JA = JAval(k)
         IB = IBval(k)
         write(*,1000) 'RUNNING', uplo, N, IA, JA, IB, 0,               &
     &                 0.0D0, 0.0D0, 0.0D0
         AM = N + IA-1
         AN = N + JA-1
         BM = N + IB-1
         if ( (AM.le.size(A,1)) .and. (AN.le.size(A,2)) .and.           &
     &        (BM.le.size(B_v,1)) ) then
            call random_number(A(IA:AM, JA:AN))
            call random_number(B_V(IB:BM))
!
!           For ease of testing, actually reflect matrix, and force it to
!           be diagonally dominant
!
            forall(i=0:N-1, j=0:N-1, i.lt.j) A(IA+i,JA+j) = A(IA+j,JA+i)
            forall (i=0:N-1) A(IA+i,JA+i) = sum(abs(A(IA:AM,JA+i)))


            if (thresh .ge. 0.0E0) then
               allocate( A0(N, N), X_V(N) )
               X_V = B_V(IB:BM)
               A0 = A(IA:AM, JA:AN)
            end if

            start = SLhpf_walltime()
            call la_posv(A(IA:AM, JA:AN), B_V(IB:BM), uplo, info)
            times = SLhpf_walltime() - start

            sres = -1.0D0
            fres = -1.0D0
            if (info .eq. 0) then
               if (thresh .ge. 0.0E0) then
                  sres = dslvcheck_v(A0, B_V(IB:BM), X_V, normA, eps)
                  if (sres .gt. thresh) then
                     stat = 'FAILED '
                     fres = dpocheck(A0, A(IA:AM, JA:AN), uplo, normA,  &
     &                               eps)
                     nfail = nfail + 1
                  else
                     stat = 'PASSED '
                     npass = npass + 1
                  end if
                  deallocate( A0, X_V )
               else
                  nbyp = nbyp + 1
                  stat = 'BYPASS '
                  sres = 0.0D0
                  fres = 0.0D0
               end if
            else
               ninfo = ninfo + 1
               stat = 'BADINFO'
               sres = dble( info )
               fres = dble( info )
            end if
         else
            stat = 'SKIPPED'
            nskip = nskip + 1
         end if
         write(*,1000) stat, uplo, N, IA, JA, IB, INFO, sres,           &
     &                 fres, maxval(times)
      end do
      write(*,2000) Nop, npass, nfail, nskip, ninfo, nbyp

 1000 format(x, a7, 2x, a1, 5I6, 2E9.1, F9.1)
 2000 format(x,I5,' TESTS:',I5,' PASSED,',I5,' FAILED,',I5,' SKIPPED,', &
     &       I5,' BAD INFO', I5, ' BYPASS.')

      RETURN
      END

      double precision function dpocheck(A, LU, uplo, normA0, eps0)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     September 1, 1998
!     Written by R. Clint Whaley, University of Tennessee, Knoxville
!
      include 'misc.h'
      real(DP), dimension(:,:), intent(in) :: A, LU
      character(len=1), intent(in) :: uplo
      real(DP), optional, intent(in) :: normA0, eps0
      integer :: n, i, j
      real(DP), allocatable, dimension(:,:) :: L, U
      real(DP) :: normA, eps
      INTERFACE
      double precision function dmatnrm(A)
         include 'misc.h'
         real(DP), intent(in), dimension(:,:) :: A
      end function dmatnrm
      END INTERFACE
!
!     Calculate norm of A and machine epsilon
!
      if (present(normA0)) then
         normA = normA0
      else
         normA = dmatnrm(A)
      end if

      if (present(eps0)) then
         eps = eps0
      else
         eps = epsilon(0.0D0)
      end if
!
!     Form L and U
!
      N = size(A, 1)
      allocate( L(n,n) )
      allocate( U(n,n) )
      L = 0
      U = 0
      forall (i=1:n) L(i,i) = 1.0D0
      if (uplo.eq.'l' .or. uplo.eq.'L') then
         forall (i=1:n, j=1:n, i .le. j) U(i,j) = LU(i,j)
         forall (i=1:n, j=1:n, i .lt. j) L(j,i) = LU(i,j)
      else
         forall (i=1:n, j=1:n, i .gt. j) L(i,j) = LU(i,j)
         forall (i=1:n, j=1:n, i .gt. j) U(j,i) = LU(i,j)
      end if
!
!     Store A - L * L' in L
!
      L = A - matmul(L,U)

      forall (i=1:N) U(i,1) = sum( L(i,:) )

      dpocheck = maxval( abs(U(:,1)) ) / (normA * N * eps)

      deallocate( L )
      deallocate( U )

      return
      end

      double precision function dslvcheck(A, x, b, normA0, eps0)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     September 1, 1998
!     Written by R. Clint Whaley, University of Tennessee, Knoxville
!
      include 'misc.h'
      real(DP), dimension(:,:), intent(in) :: A, x, b
      real(DP), optional, intent(out) :: normA0, eps0
      integer :: n, nrhs, i, j
      real(DP), allocatable, dimension(:,:) :: Ax_b
      real(DP), allocatable, dimension(:) :: work(:)
      real(DP) :: normA, normx, eps, num, den, nrm, maxnrm
      INTERFACE
      double precision function dmatnrm(A)
         include 'misc.h'
         real(DP), intent(in), dimension(:,:) :: A
      end function dmatnrm
      END INTERFACE

!
!     Calculate norm of A and machine epsilon
!
      normA = dmatnrm(A)
      eps = epsilon(0.0D0)

      n = size(A, 1)
      nrhs = size(x, 2)
      allocate( Ax_b(n, nrhs) )
      allocate( work(n) )
!
!     Check solve: ||Ax - b|| / ( ||A|| * ||x|| * N * eps)
!
      maxnrm = 0.0
      den = normA * eps * N
      Ax_b = matmul(A, x) - b
      do i = 1, nrhs
         work(:) = abs( Ax_b(:,i) )
         num = maxval( work(:) )
         work(:) = abs( x(:,i) )
         nrm = num / ( den * maxval( work(:) ) )
         if (nrm .gt. maxnrm) maxnrm = nrm
      end do

      dslvcheck = maxnrm

      deallocate( Ax_b )
      deallocate( work )

      if (present(normA0)) normA0 = normA
      if (present(eps0)) eps0 = eps

      return
      end



      double precision function dslvcheck_v(A, x_v, b_v, normA0, eps0)
!
!  -- Layer 1 ScaLAPACK HPF wrapper routine, (version 1.1) --
!     Written by C. A. Papadopoulos, University of Tennessee, Knoxville
!     Adapted from SLhpf routines written by R. Clint Whaley,
!     University of Tennessee, Knoxville
!     September 1, 1998
!
      include 'misc.h'
      real(DP), dimension(:,:), intent(in) :: A
      real(DP), dimension(:), intent(in) :: x_v, b_v
      real(DP), optional, intent(out) :: normA0, eps0
      integer :: n, i, j
      real(DP), allocatable, dimension(:) :: Ax_b
      real(DP) :: work
      real(DP) :: normA, normx, eps, num, den, nrm, maxnrm
      INTERFACE
      double precision function dmatnrm(A)
         include 'misc.h'
         real(DP), intent(in), dimension(:,:) :: A
      end function dmatnrm
      double precision function dmatnrm_v(V)
         include 'misc.h'
         real(DP), intent(in), dimension(:) :: V
      end function dmatnrm_v
      END INTERFACE

!
!     Calculate norm of A and machine epsilon
!
      normA = dmatnrm(A)
      eps = epsilon(0.0D0)

      n = size(A, 1)
      allocate( Ax_b(n) )
!
!     Check solve: ||Ax - b|| / ( ||A|| * ||x|| * N * eps)
!
      maxnrm = 0.0
      den = normA * eps * N
!
! note since pghpf matmul does not work on vectors use a double for loop
!      Ax_b = matmul(A, x_v) - b_v
        Ax_b = 0.0
        do i=1,n
           do j=1,n
              Ax_b(i) = Ax_b(i) + A(i,j)*x_v(j)
           end do
        end do
        Ax_b = Ax_b - b_v

      work = maxval(abs(Ax_b))
      num =  work
      work = maxval(abs(x_v))
      nrm = num / ( den *  work )
      if (nrm .gt. maxnrm) maxnrm = nrm

      dslvcheck_v = maxnrm

      deallocate( Ax_b )

      if (present(normA0)) normA0 = normA
      if (present(eps0)) eps0 = eps

      return
      end

